/***************************************************************************
                          components.cpp  -  description
                             -------------------

    functionality        : for constructing and handling component matrices
    relevant classes     : component_matrix
    relevant functions   : get_partition

    begin                : Thu Dec 6 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "components.h"
#include "trees.h"
#include <cmath>
#include <iostream>


class and_tree;

void component_matrix ::fill_matrix_cond(int ActualArg, condition ActualCond, int depth)
{
  int k;
  k=get_partition(ActualCond);
  if (k!=-1)
    V[NumAttributes*k+ActualArg]+= pow(0.5,depth);  //1/float(depth+1);
}

void component_matrix ::fill_matrix_rule(rule Rule, int depth)
{
  unsigned int i,j;
  for (i=0;i<Rule.V_condition.size() -1 ;i++)
     for (j=0;j<Rule.V_condition[i].size();j++)
       fill_matrix_cond(i, Rule.V_condition[i][j], depth);
}


void component_matrix ::fill_matrix(vector <and_tree*> nodes)
{
unsigned int i;
for (i=0;i<nodes.size();i++)
  fill_matrix_rule(nodes[i]->Rule,nodes[i]->Depth);
}

void component_matrix::Add(const component_matrix &C) {
  for(int i=0; i < NumSplits*NumAttributes; i++) {
     V[i]+= C.V[i];
  }
}

void component_matrix::Reset() {
  for(int i=0; i < NumSplits*NumAttributes; i++) {
    V[i] = 0;
  }
}


// puts the f% of elements to 1 and the rest to 0 randomly
void component_matrix::Randomise1(float f) {
  Reset();
  int Tam= NumSplits*NumAttributes;
  int Num= (int)(f*Tam);
  if (Num > Tam)
    Num= Tam;
  for(int i=0; i<Num;) {
    int Pos= (int)RandUniform(0, Tam);
    if (V[Pos] == 0.0) {
      V[Pos]= 1.0;
      i++;
    }
  }
}


void component_matrix ::Show() {
    const int ColW= 5;
    int i;
    cout << "P\\R" << setw(ColW);
    for(i=0; i < NumAttributes; i++) {
     cout << "  C" << setw(ColW-3) << i << "|" ;
   }
    for(i=0; i < NumAttributes*NumSplits; i++) {
       if ((i % NumAttributes) == 0)
          cout << endl << "  C" << setw(ColW-3) << i / NumAttributes;
       cout << setw(ColW) << V[i] << "|";
      }
   cout << endl;
  }

  double component_matrix ::EuclideanDistance(const component_matrix &m2) {
    double Dist= 0.0;
    for(int i=0; i < NumSplits*NumAttributes; i++) {
      Dist += Sqr(V[i] - m2.V[i]);
   }
   return sqrt(Dist);
 }

   double component_matrix ::ManhattanDistance(const component_matrix &m2) {
    double Dist= 0.0;
    for(int i=0; i < NumSplits*NumAttributes; i++) {
      Dist += abs(V[i] - m2.V[i]);
   		}
   return Dist;
 	}
 	
 	
