/***************************************************************************
                          criteria.cpp  -  description
                             -------------------

    functionality        : for computing some criteria and defines structures related to these options
    relevant classes     : split_partition, split_selection_criteria ...
    relevant functions   :

    begin                : Tue Oct 23 2001
    copyright            : (C) 2001 by C. Ferri, Jos Hernndez-Orallo
    email                : {cferri, jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "trees.h"
#include "criteria.h"
#include "utils.h"
#include "cost.h"
#include "roc.h"
#include <cmath>
#include <vector>
// #define TREE_DEBUG

// Computes the different criteria according to the values: SSC, ...
void Compute_Criteria_AND(and_tree &AT, const dataset &Train, const options &Options ) {
      int nC= Train.NumClasses();

     float Card= AT.Cardinality;

    vector<double> StratifiedDistribution;

    if (Card!= 0) {
       if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
           (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING)) {
        vector<double> Dist(nC);
        double Acc= 0;
         for(int j=0; j < nC; j++) {
           double jValue;
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION)
              jValue= AT.SmoothDistribution[j];
            else
              jValue= AT.ClassDistribution[j];
           Dist[j]= jValue * Options.WeightVector[j];
           Acc+= Dist[j];
         }
         if (Acc) {
           for(int j=0; j < nC; j++) {  // Normalitzar
             Dist[j]/=Acc;
           }
         }
        StratifiedDistribution= Dist;
      }
    }

    if ((Options.SSC == GAIN) || (Options.SSC == GAIN_RATIO) || (Options.SSC == C45))  {
      float IE = 0;

      if (Card != 0) {
         for(int j=0; j < nC; j++) {
          float pEj;
          if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
           pEj= (float)AT.SmoothDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
           pEj= (float)AT.ClassDistribution[j] / Card;
          else if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
                    (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING))
           pEj= StratifiedDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_COSTS)
            pEj= (float)AT.CostProb[j];
          if (pEj != 0.0)
            IE+= pEj * log2(pEj);
        }
      }
      AT.Crit.Gain.InfoE = -IE;
        #ifdef TREE_DEBUG
          cout << "InfoE of Node AND #" << AT.Num << " computed: " << AT.Crit.Gain.InfoE << endl;
        #endif

      if  ((Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
        // nothing special to do
      }
    }
    else if ((Options.SSC == CART) || (Options.SSC == DKM)) {
      float GINI = 1;
      if (Card != 0) {
         for(int j=0; j < nC; j++) {
          float pEj;
          if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
           pEj= (float)AT.SmoothDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
           pEj= (float)AT.ClassDistribution[j] / (float)Card;
           else if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
                 (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING))
           pEj= StratifiedDistribution[j];
           else if (Options.ProbabilityInSplittingCriteria == FROM_COSTS) {
            pEj= (float)AT.CostProb[j];
          }
          GINI+= Sqr(pEj);
        }
        GINI= 1 - GINI;
      }
     if (Options.SSC == CART) {
      AT.Crit.Cart.GiniE = GINI;
     }
     else { // if (Options.SSC == DKM) {
       AT.Crit.Cart.GiniE = pow(GINI, 1/float(nC));
      }

        #ifdef TREE_DEBUG
          cout << "GiniE of Node AND #" << AT.Num << " computed: " << AT.Crit.Cart.GiniE << endl;
        #endif
    }

    else if (Options.SSC==DESC_MDL)	{
   	
     	unsigned int j,k;
 	    indext h;
 	    vector<vindex> tix ;
      for (j=0;j<Train.num_arguments;j++)  // atenci que axi els arguments tamb tenen
   																		 //la classe
       {
         k=Train.exetable[j].size(); // numero de indexs diferents per a eixe arg o classe
     	  vindex vix =vindex(k);
       	tix.push_back(vix);
       }

	    for (h=0;h<AT.Covered.size();h++)// atenci que aci els arguments tamb tenen la classe
  		  for (j=0;j<Train.num_arguments;j++)
  		  {if (Train.KindOf(j)==DISCRETE_KIND)
			  tix[j].set_index(Train.Nth_arg(AT.Covered[h],j).Value_d);}
         AT.Crit.GMDL.cost=compute_cost(Train,tix,Train.exetable,AT.Cardinality);

         // I QUE PASSA AMB NUMERIC_KIND I IGNORE_KIND
    }
    else if (Options.SSC==SPLIT_EXPECTED_ERROR) {
     // Nothing to do
    }
    else if (Options.SSC==LOCAL_ROC_AREA) {
     // Nothing to do
    }
    else if (Options.SSC==MSE_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==MSEE_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==MCE_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==LOGLOSS_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==MIXTURE_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==AUCH_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==AUCS_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==MGINI_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==GENENTROPY_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==SQDIF_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==ROCV_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==ONE_POINT_LOCAL_ROC_AREA) {
     // Nothing to do
    }
    else if (Options.SSC==LEFT_FIRST) {
     // Nothing to do
    }
    else {
      cout << "Unknown Split criterion in function Compute_Criteria_AND" << endl;
   }
 }



// #define SMOOTH(a,b,c)   ((c)?(((a)+1)/(b+1)):(((a))/(b)))


inline double SMOOTH(double a, double b, double nNodes, const options &Options) {
  if (Options.NodeSmoothing.Method == NO_SMOOTHING) {
    return a/b;
  }
  else if (Options.NodeSmoothing.Method == LAPLACE) {
    return (a+1)/(b+nNodes);  // Authentic Laplace Smoothing
  }
  else if (Options.NodeSmoothing.Method == K_ESTIMATE) {
    double k= Options.NodeSmoothing.m;
   // k= 1000;
    return (a+k)/(b+k);  // This is a smoothing not generally used
  }
  else if ((Options.NodeSmoothing.Method == M_ESTIMATE) || (Options.NodeSmoothing.Method == M_ESTIMATE_UNIFORM)) {
    double m= Options.NodeSmoothing.m;
   // m= 1000;
    return (a+m*(1/nNodes)) / (b+m);
  }
  else {
    cout << "Node Smoothing Method Unknown" << endl;
    return 0;
  }
}


bool GLOBALFirstTime= true;
double GLOBALCorrection= 1;

// corrects some criteria (no. of children).
void Split_Children_Correction(or_tree &OT, const dataset &Train, const options &Options, long nC, double NumChildren) {

      and_tree *Parent= (and_tree *) OT.Parent;
      double CardParent= (double)Parent->Cardinality; // cardinality of all the splits
/*
      CardMean = pow(CardMean, 1/double(NumChildren));

      double Corr= 1;
      if (CardParent)
        Corr= NumChildren * (CardMean / CardParent);

      OT.Split_Optimality*= pow(Corr, 1 / double(nC));
*/


        double Correction;

//      if (false) {
//      if (true) {
      if (Options.SplittingCriteriaChildrenCorrection) {


        double degrees_of_freedom= (Train.Number_Of_Nominal()
                                    // + Train.Number_Of_Numeric()
                                    ) / nC;  // no. of nominal attributes / no. of classes

      //  degrees_of_freedom *= 2; // (when there are just two classes it should be the number of attributes)

        double normalised_sample_freedom;
        double cardi= CardParent; // MinCard * NumChildren * 2;
        if (cardi > 0) {
          double sample_freedom= degrees_of_freedom / cardi; ;
          normalised_sample_freedom= 1 - exp(-sample_freedom);
        }
        else {
          normalised_sample_freedom= 1;  // maximum importance to the correction
        }

        // if (NumChildren > 2)
        //  cout << "no. of children" << NumChildren << endl;

        normalised_sample_freedom= pow(normalised_sample_freedom, 3.0);  // to make less effect

        double ProbMin;

        // ProbMin= MinCard / double(CardParent);

        ProbMin *= 2;   // Now it will be 1 when it is evenly distributed or near 0 when not
        // if it is close to 1 then we shouldn't take children into account.

        // Hence
        normalised_sample_freedom= 1 - ProbMin;

        normalised_sample_freedom = normalised_sample_freedom * 2.0 / nC;

        normalised_sample_freedom= pow(normalised_sample_freedom, 12.0);


    //    normalised_sample_freedom= long(CardParent) % 2;  // random
        normalised_sample_freedom= 1;

        Correction=  pow(NumChildren, normalised_sample_freedom * Options.SplittingCriteriaChildrenCorrectionFactor);

         //  Correction=  1 + pow(NumChildren, 1.0 / 10.0);

        // Correction= 1 + log(double(NumChildren));


        // Correction= NumChildren;

       // Correction= pow(Correction, 1.0/pow((nC - 1.0),10));  // for two classes it makes no effect


        // Correction= pow(Correction, -(nC - 3.0) );  // for two classes it makes no effect


        // Correction= log(double(NumChildren))  * normalised_sample_freedom;

/*
        // NOVA RATLLADA
        double AccDiff= 0;
        if (CardParent != 0) {
          list<tree *>::iterator It, ItEnd;
          ItEnd= OT.Children.end();
          for(It= OT.Children.begin(); It != ItEnd; It++) {
            and_tree *Child= (and_tree *)*It;

            // float p= Child->Cardinality / (float)CardParent;
            double Diff= abs(long(Child->ClassDistribution[0] - Child->ClassDistribution[1]));
            // for(int i= 0; i < nC; i++) {
            AccDiff+= Diff;
          }
        }

        double Quot= CardParent / AccDiff;

        Correction=  pow(NumChildren, Quot);
*/

        double Min= CardParent;
        double Max= 0;
        double Mean= CardParent / double(nC);
        double Stdev= 0;
        double SqrSum= 0;
        for(int i= 0; i < nC; i++) {
          double Ara= Parent->ClassDistribution[i];
          if (Ara < Min) {  // ABANS ESTAVA CardParent
            Min= Ara;
          }
          if (Ara > Max) {
            Max= Ara;
          }

          SqrSum+= Ara * Ara;

        }
        Stdev= nC * SqrSum - pow(CardParent, 2);
        Stdev= sqrt(Stdev / (nC * (nC - 1.0)) );

        // Stddev is now the standard deviation of the cardinalities of all the classes

        double RatioStdev= Stdev / CardParent;
        double RatioMin= Min / CardParent;
        double RatioMax= Max / CardParent;

        if (GLOBALFirstTime) {

          // TAXATIVE
          if (RatioMin > 0.35) {  // 0.35
            Correction= 1;
            GLOBALCorrection= 0;
          }
          else {
            Correction= pow(NumChildren, 15);
            GLOBALCorrection= 15;
          }

          // SMOOTHER

   //       GLOBALCorrection= 15 * pow(2 * Ratio - 1, 2);  // BONA
   //       GLOBALCorrection= 15 * pow(2 * Ratio - 1, 3);  // BONA I PROVADA
        //  GLOBALCorrection= pow(2 * Ratio - 1, 1);  // similar al anterior
      //    GLOBALCorrection= pow(2 * Ratio - 1, 2.0/3.0);

           GLOBALCorrection= 2 * pow(2 * RatioMin - 1, 2);  // MILLOR FINS ARA I PROVADA AMB CARDPARENT
           // PROVAR!!! esta que es molt similar i ms senzilla:
         //  GLOBALCorrection= 1 * fabs(pow(Ratio - 1/2.0, 1));
           // o b
      //     GLOBALCorrection= 2 * pow(Ratio - 1/2.0, 1);

      //    GLOBALCorrection= 1 * fabs(pow((2 * Ratio - 1)/nC, nC));


  //        GLOBALCorrection= 2 * fabs(pow(2 * RatioMax - 1, nC));

          GLOBALCorrection= abs(4-nC) * (3 - nC)  * fabs(pow(RatioStdev, 2));  // LA MILLOR FINS ARA.

//         GLOBALCorrection= (3.5 - nC)  * fabs(pow(RatioStdev, 2));  // LA MILLOR FINS ARA.



  //        GLOBALCorrection= RatioStdev;

          // theoreticyally GOOD FOR MORE THAN 2 CLASSES;
         //  Ratio= 1 - Ratio;
         //  GLOBALCorrection= 1 * fabs(2* pow((1 - 2 * Ratio), nC));


/*
      //    GLOBALCorrection= 1/2.0 * pow(2 * Ratio - 1, 1);

          GLOBALCorrection= (4 - nC) * pow(2 * Ratio - 1, 2);  // Extensi no. de classes
                                                               // NO VA MOLT BE


          GLOBALCorrection= 2 * pow(2 * Ratio - 1, nC);  // Extensi no. de classes
                                                               // NO VA MOLT BE

          GLOBALCorrection= 2 * pow(nC * Ratio - 1, nC);  // Extensi no. de classes
                                                               // NO VA MOLT BE

          GLOBALCorrection= 2 * (3 - nC) * pow(2 * Ratio - 1, nC);  // Extensi no. de classes
                                                               // NO VA MOLT BE
*/
        //  GLOBALCorrection= 2; // -1.0/1.0;



          Correction= pow(NumChildren, GLOBALCorrection);


          GLOBALFirstTime= false;
        }
        else
          Correction= pow(NumChildren, GLOBALCorrection);

        // Correction=1;

       // Correction= 1 + pow(NumChildren - nC, GLOBALCorrection);
        //  Correction= 1 + fabs(pow(1+NumChildren - nC, 3));
        // Correction= pow(nC, 4) + fabs(pow(NumChildren - nC, 4));  // VA B PER A MS DE 2 CLASSES

        Correction= (pow(fabs(NumChildren), GLOBALCorrection));  // LA MILLOR

//        Correction= nC + pow( fabs(NumChildren - nC), GLOBALCorrection);


/*
        // GainRatio Correction
        float AS= 0;
        double Min= 10000000000.0;
        if (CardParent != 0) {
          list<tree *>::iterator It, ItEnd;
          ItEnd= OT.Children.end();
          for(It= OT.Children.begin(); It != ItEnd; It++) {
            and_tree *Child= (and_tree *)*It;
            #ifdef TREE_DEBUG
            if (Child == NULL)  {
              cout << "Child NULL en Compute_Criteria de or_node";
            }
            #endif
            // float p= Child->Cardinality / (float)CardParent;
            float p= SMOOTH(Child->Cardinality, (float)CardParent, OT.Children.size(), Options);
            if (p!=0) {
              AS+= p * log2(p);
            }

            if (Child->Cardinality < Min)
              Min= Child->Cardinality;
          }
        }
        Correction = -AS;  // GainRatio

        if (Min != 0) {
          Correction= CardParent / Min;
        }
        else
          Correction= 20000000L;
*/

        if (Options.SplittingCriteriaChildrenCorrectionMean) {
          OT.Crit.Gain.GainET= OT.Split_Optimality; // before correction (to be used as C4.5 criterion. AT LEAST THE MEAN)
        }

        OT.Split_Optimality /= Correction;
      }
}

// Computes different criteria according to values: SSC, ...
void Compute_Criteria_OR(or_tree &OT, const dataset &Train, const options &Options ) {

    int nC= Train.NumClasses();

    if (Options.SSC == LEFT_FIRST) {

      // Cannot be implemented locally. We don't know whether it is the first
      // Nothing to do here then.
      // IT IS IMPLEMENTED AND IT WORKS IN FUNCTION and_tree::Select_Best_OR_For_Split

    }
    else if ((Options.SSC == GAIN) || (Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
      float G = ((and_tree *)OT.Parent)->Crit.Gain.InfoE;
      indext Card = ((and_tree *)OT.Parent)->Cardinality;
      float AG= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          #ifdef TREE_DEBUG
          if (Child == NULL)  {
            cout << "Child NULL en Compute_Criteria de or_node";
          }
          #endif
         AG+= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options) * Child->Crit.Gain.InfoE;
          // AG+= (Child->Cardinality / (float)Card) * Child->Crit.Gain.InfoE;
        }
      }
      OT.Crit.Gain.GainET = G - AG;
      if (Options.SSC == GAIN) {
         OT.Split_Optimality = OT.Crit.Gain.GainET;
      }
      else if  ((Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
        float AS= 0;
        if (Card != 0) {
          ItEnd= OT.Children.end();
          for(It= OT.Children.begin(); It != ItEnd; It++) {
            and_tree *Child= (and_tree *)*It;
            #ifdef TREE_DEBUG
            if (Child == NULL)  {
              cout << "Child NULL en Compute_Criteria de or_node";
            }
            #endif
            // float p= Child->Cardinality / (float)Card;
            float p= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options);
            if (p!=0) {
              AS+= p * log2(p);
            }
          }
        }
        OT.Crit.Gain.SplitET = -AS;
        if (AS != 0) {
          OT.Crit.Gain.GainRatioET = OT.Crit.Gain.GainET / OT.Crit.Gain.SplitET;
        }
        else {
          OT.Crit.Gain.GainRatioET = OT.Crit.Gain.GainET;
        }
        OT.Split_Optimality = OT.Crit.Gain.GainRatioET;
      }
    }
    else if ((Options.SSC == CART) ) { // || (Options.SSC == DKM)) {
      float G = ((and_tree *)OT.Parent)->Crit.Cart.GiniE;
      indext Card = ((and_tree *)OT.Parent)->Cardinality;
      float AG= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          #ifdef TREE_DEBUG
          if (Child == NULL)  {
            cout << "Child NULL en Compute_Criteria de or_node";
          }
          #endif
          // float p= Child->Cardinality / (float)Card;
          double p= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options);
          AG+= p * Child->Crit.Cart.GiniE;
        }
      }
      OT.Crit.Cart.GainET = G - AG;
      OT.Split_Optimality = OT.Crit.Cart.GainET;
    }
    else if (Options.SSC==DESC_MDL) {
	
	    float AC=0;
      AC+= calcula_partv(((and_tree *)OT.Parent)->OpenVariables);
	    list<tree *>::iterator It, ItEnd;
	    indext Card = ((and_tree *)OT.Parent)->Cardinality;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          AC+= (Child->Crit.GMDL.cost);
        }
        OT.Crit.GMDL.cost = AC;
        OT.Split_Optimality =1/AC;
      }
	  }
    else if (Options.SSC==SPLIT_EXPECTED_ERROR) {
      OT.Split_Optimality= -OT.BackupError;  // WITH JUST THIS, IT WORKS WELL

      int nC= Train.NumClasses();
      long NChildren= OT.Children.size();

      vector<double> p(nC);  // maxims
      vector<double> Accum(nC);  // accum
      int j;
		  for(j=0; j < nC; j++) {
		    Accum[j]= 0;
      }


      list<tree *>::iterator It, ItEnd;

      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> q(nC);
        long Card= Child->Cardinality;

        double First= 0, Second= 0;
        int Firstj=0, Secondj= 0;

        if (Card != 0) {
          int j;
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
		      for(j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
/*
            if (j != Child->PredictedClass) {
              Accum[j]+= pNode * fabs(pow(p[j], 2.0));
            }
*/
             //          Accum[j]+=   -    pNode * log(p[j] / (1-p[j]));

            if (First <= p[j]) {
              Second= First;
              Secondj= Firstj;
              Firstj= j;
              First= p[j];
            }
          }

          Accum[0]+= - pNode * Sqr(fabs(First-Second));

        }

      }

      double AcVar= 0;

		  for(j=0; j < nC; j++) {
		    AcVar+= Accum[j];  // ignoring class distribution
      }

      // Children correction
		  // AcVar = AcVar / double(NChildren); //   / double(NChildren * (NChildren - 1.0));
		  // AcVar = AcVar * double(NChildren); //   / double(NChildren * (NChildren - 1.0));

		
      OT.Split_Optimality= -AcVar;  // more difference worse
    }	
    else if (Options.SSC==LOCAL_ROC_AREA) {
       double Area;

       vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }


       vector<vector<double> > Points;
       compute_ROC_points(Vector, Points, ROC_WITH_TRAIN, Options,0,1);
       compute_ROC_AREA(Points, Area);
   //    cout << "ROC Area: " << Area << endl;
       OT.Split_Optimality= Area;
    }
/*  ESTE ES EL L'ANTIC MSE QUE NO PERMET SMOOTHING
    else if (Options.SSC==MSE_SPLIT) {
      double MSE1;

      vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      MSE1= compute_MSE(Vector, Options, MSE_WITH_TRAIN);
      MSE1= MSE1/Train.NumClasses();
      MSE1= 1 - MSE1;

      OT.Split_Optimality= MSE1;
    }
*/

/*
    else if (true) {
                // Options.SSC==MSE_BIS_SPLIT) {

      double MSE=0;
      int nC= Train.NumClasses();
      double TotalTestValue= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        double MSE0= 0;

        for(int j=0; j<nC; j++) {
            double ProbTrain;
            if (Card == 0) {  // We assume equal probability for all the classes
              ProbTrain = 1 / double(nC);
            }
            else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain= (float)Child->SmoothDistribution[j];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain= (float)Child->ClassDistribution[j] / (float)Card;
            }

            // MSE0+= ProbTrain - Sqr(ProbTrain);
            // MSE0+= ProbTrain + Sqr(ProbTrain) - 2*pow(ProbTrain, 3);

            // MSE0+= ProbTrain * pow(fabs(1 - ProbTrain), 3) + (1 - ProbTrain) * pow(ProbTrain, 3);

           // MSE0+= Sqr(ProbTrain);

            MSE0+= ProbTrain * pow(fabs(1 - ProbTrain), 2);

        }

        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
         double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

       // MSE+= pow(pNode, 2) * MSE0;
        MSE+= pNode * MSE0;

      }
      OT.Split_Optimality= 1 - MSE;
    }
*/


    else if (Options.SSC==MSEE_SPLIT) {
      double MSE=0;
      int nC= Train.NumClasses();
      double TotalTestValue= 0;

      int NumChildren= OT.Children.size();


      vector<double> pparent(nC);

		  for(int j=0; j < nC; j++) {
		    bool LOCAL= true;
		    bool CONSTANT= false;
		    if (CONSTANT) {
		     // pparent[j]= 0.5;
		      pparent[j]= 1 / double(nC);
		    }
		    else
		      if (LOCAL) {
		        and_tree *Par= (and_tree *)OT.Parent;
//            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              pparent[j]= (Par)->SmoothDistribution[j];
//            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
//              pparent[j]= (float)(Par)->ClassDistribution[j] / (float)Par->Cardinality;
		      }
		      else {  // Global. Everything computed wrt. global distribution
		        pparent[j]= (Train.ClassDistribution()[j] + 1.0) / (float)(Train.Cardinality() + nC);
		        //  pparent[j]*= nC / 2.0;
		        //  pparent[j] = sqrt(pparent[j]);
			    }
		  }

/*		  for(int j=0; j < nC; j++) {
		    pparent[j]= (pparent[j] + 0.5) / 2.0;
      }
*/
      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        double MSE0= 0;

        for(int j=0; j<nC; j++) {
          long TestValue;
          TestValue= Child->ClassDistribution[j];

/*
          TestValue++; // DOUBLE SMOOTHING!!  // A CURIOSAMENT EMPITJORA ELS RESULTATS
          double pTestValue=0.0;
          // if (Card > 0)
            pTestValue = (TestValue) / double(Card + nC);  // don't sum 1 to the numerator. TestValue has already been incremented
*/
/*
          const int K= 1;
          TestValue+= K; // K SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + K);
*/

          // WITHOUT DOUBLE SMOOTHING
          double pTestValue=0.0;
          if (Card > 0) {
            pTestValue = TestValue / double(Card);
//            pTestValue= (float)Child->SmoothDistribution[j];
          }
          else {
            pTestValue= 1 / double(nC);  // SENSE AIX VA UN PELIU MILLOR
          }

/*
          double alpha= Card / double(Train.Cardinality());
          alpha= max(0.2 - alpha, 0.0);

          pTestValue= ((1 - alpha)* pTestValue + alpha * pparent[j]); // The tendency will be to also tend a bit to the parent distribution
*/
/*
          MSE0+= pTestValue * pTestValue * Sqr(1-pTestValue);
          MSE0+= pTestValue * fabs(1 - pTestValue) * Sqr(pTestValue);
*/

     //

          for(int k=0; k<nC; k++) {

            double ProbTrain;
            if (Card == 0) {  // We assume equal probability for all the classes
              ProbTrain = 1 / double(nC);
            }
            else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain= (float)Child->SmoothDistribution[k];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain= (float)Child->ClassDistribution[k] / (float)Card;
            }

//            double Smoothed0= 1.0/(Card + nC);
            if (j!=k) {
              // MSE0+= pow(pTestValue,1) * pow(ProbTrain, 3); // Sqr(0-ProbTrain);
            //  MSE0+= pow(pTestValue,1) * ProbTrain * fabs(pow(ProbTrain, 2)); // Sqr(0-ProbTrain);
//              MSE0+= pTestValue * Sqr(0-ProbTrain);
            //  MSE0+= pTestValue * Sqr( (0-ProbTrain) / (0 - pparent[j]));
            //  MSE0+= pTestValue * Sqr((0-ProbTrain));

            //   MSE0+= pTestValue * Sqr((0-ProbTrain));

             // MSE0+= pTestValue * pow(2.0, fabs(0-ProbTrain));

             //  MSE0+= pTestValue * pow(fabs(Smoothed0-ProbTrain), 2);
/*
              for(int l=0; l<nC; l++) {

                double ProbTrain2;
                if (Card == 0) {  // We assume equal probability for all the classes
                  ProbTrain2 = 1 / double(nC);
                }
                else {
                  if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                    ProbTrain2= (float)Child->SmoothDistribution[l];
                  else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                    ProbTrain2= (float)Child->ClassDistribution[l] / (float)Card;
                }

                MSE0+= pTestValue * (ProbTrain) * Sqr(0-ProbTrain);
              }
*/
             //              MSE0+= pTestValue * (ProbTrain) * Sqr(0-ProbTrain);
                MSE0+= pTestValue * (1 - pTestValue) * fabs(pow(0-ProbTrain, 2));
            }
            else {
              // MSE0+= pow(pTestValue,1) * pow(1-ProbTrain, 3); // Sqr(1-ProbTrain);
        //     MSE0+= pow(pTestValue,1) * ProbTrain * fabs(pow(1-ProbTrain, 2)); // Sqr(1-ProbTrain);
//              MSE0+= pTestValue * Sqr(1-ProbTrain);
         //     MSE0+= pTestValue * Sqr((1-ProbTrain) / (1 - pparent[j]));
        //      MSE0+= pTestValue * Sqr((1-ProbTrain)/nC);
        //      MSE0+= pTestValue * fabs(1 - Sqr(ProbTrain));
          //   MSE0+= pTestValue * pow(2.0, fabs(1-ProbTrain));

          //              double Smoothed1= 1 - Smoothed0;
          //              MSE0+= pTestValue * pow(fabs(Smoothed1-ProbTrain), 2);

             //              MSE0+= pTestValue * pTestValue * Sqr(1-ProbTrain);
              MSE0+= pTestValue * (1 - pTestValue) * fabs(pow(1-ProbTrain, 2));

            }
          }

          TotalTestValue+= TestValue;
        }

        Child->LocalOptimality= 1 - MSE0;  // NO T (I NO POT TINDRE) CHILDREN CORRECTION


        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
         double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

       // MSE+= pow(pNode, 2) * MSE0;
        MSE+= pNode * MSE0;

      }

      MSE/= /* TotalTestValue * */ nC;
      // MSE/= TotalTestValue * nC;             // LA DIFERENCIA ES INAPRECIABLE. ABANS ESTAVA LO DE DALT...

      OT.Split_Optimality= 1 - MSE;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }


    else if ((Options.SSC==MSE_SPLIT) || (Options.SSC==MCE_SPLIT)) {
      double MSE=0;
      double TotalTestValue= 0;

      double CardMean= 1;
      double MinCard= -1;

      int NumChildren= OT.Children.size();
      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        if ((Card < MinCard) || (MinCard == -1)) {
          MinCard= Card;
        }

        CardMean*= Card;

        double MSE0= 0;

        for(int j=0; j<nC; j++) {
          long TestValue;
          TestValue= Child->ClassDistribution[j];

/*
          TestValue++; // DOUBLE SMOOTHING!!  // A CURIOSAMENT EMPITJORA ELS RESULTATS
          double pTestValue=0.0;
          // if (Card > 0)
            pTestValue = (TestValue) / double(Card + nC);  // don't sum 1 to the numerator. TestValue has already been incremented
*/
/*
          const int K= 1;
          TestValue+= K; // K SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + K);
*/

          // WITHOUT DOUBLE SMOOTHING
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card);
          else {
            pTestValue= 1 / double(nC);  // AMB AIX NO S'HA PROVAT
          }

          for(int k=0; k<nC; k++) {

            double ProbTrain;
            if (Card == 0) {  // We assume equal probability for all the classes
              ProbTrain = 1 / double(nC);
            }
            else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain= (float)Child->SmoothDistribution[k];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain= (float)Child->ClassDistribution[k] / (float)Card;
            }

            if (j!=k) {
              if (Options.SSC==MCE_SPLIT) {
                MSE0+= pow(pTestValue,1) * pow(ProbTrain, 3); // Sqr(0-ProbTrain);
              }
              else if (Options.SSC==MSE_SPLIT) {
                MSE0+= pTestValue * Sqr(0-ProbTrain);
              }
            }
            else {
              if (Options.SSC==MCE_SPLIT) {
                MSE0+= pow(pTestValue,1) * pow(1-ProbTrain, 3); // Sqr(1-ProbTrain);
              }
              else if (Options.SSC==MSE_SPLIT) {
                MSE0+= pTestValue * Sqr(1-ProbTrain);
              }
            }
          }
          TotalTestValue+= TestValue;
        }

        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
         double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

       // MSE+= pow(pNode, 2) * MSE0;
        MSE+= pNode * MSE0;

      }

      MSE/= /* TotalTestValue * */ nC;
      // MSE/= TotalTestValue * nC;             // LA DIFERENCIA ES INAPRECIABLE. ABANS ESTAVA LO DE DALT...

      OT.Split_Optimality= 1 - MSE;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
    }


    else if (Options.SSC==LOGLOSS_SPLIT) {
      double MSE;

      vector<and_tree *> Vector;

      int NumChildren= OT.Children.size();

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      MSE= compute_LogLoss(Vector, Options, MSE_WITH_TRAIN);
      // MSE= MSE/2;
      MSE= 1 - MSE;

      OT.Split_Optimality= MSE;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }
    else if (Options.SSC==AUCH_SPLIT) {
      double AUCH;

      vector<and_tree *> Vector;

      int NumChildren= OT.Children.size();

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      AUCH=compute_AUC_by_Hand(Vector,ROC_WITH_TRAIN,Options);

      OT.Split_Optimality= AUCH;


   //   OT.Split_Optimality /= pow(NumChildren, 1/15.0);

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }
    else if (Options.SSC==AUCS_SPLIT) {
      double AUCS;

      vector<and_tree *> Vector;

      int NumChildren= OT.Children.size();

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

//      AUCS=compute_AUC_by_Srinivasan(Vector,ROC_WITH_TRAIN,Options, true);  // last argument use or not probability
      AUCS=compute_AUC_by_Ploro(Vector,ROC_WITH_TRAIN,Options, false);  // last argument use or not probability

      OT.Split_Optimality= AUCS;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
    }
    else if (Options.SSC==CHILDVSCHILD_SPLIT) {
      int nC= Train.NumClasses();
      long NChildren= OT.Children.size();

      vector<double> Sum(nC);
      vector<double> SumSq(nC);
      vector<double> p(nC);

		  for(int j=0; j < nC; j++) {
		    Sum[j]= 0;
		    SumSq[j]= 0;
		  }

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> q(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
          // la prob del node no es gasta...

          int j;
		      for(j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;

            p[j]-= 0.5; // Now it goes from -0.5 to +0.5
            p[j]*= pNode;

            Sum[j]+= p[j];
            SumSq[j]+= Sqr(p[j]);
          }
        }
      }

      double AcVar= 0;
      int j;
		  for(j=0; j < nC; j++) {
		    double Var= NChildren * SumSq[j] - Sqr(Sum[j]);
		
		    Var= Var / double(NChildren);
		    // Var= Var / (double(NChildren) * (NChildren - 1.0));
		
		    // Var= Var * ((and_tree *)OT.Parent)->SmoothDistribution[j];      // ponderar amb la probabilitat de cada classe en el pare
		
		    // Var = sqrt(Var); // sumar desviacions en compte de les varianes
		
		    AcVar+= Var;
      }

      OT.Split_Optimality= AcVar;  // more variance better
    }


    else if (Options.SSC==CHILDVSPARENT_SPLIT) {
      int nC= Train.NumClasses();
      long NChildren= OT.Children.size();

      vector<double> SumSq(nC);
      vector<double> p(nC);
      vector<double> pparent(nC);
      vector<double> mu(nC);  // mitges
      double Mean= 0;

		  for(int j=0; j < nC; j++) {
		    SumSq[j]= 0;
		    mu[j]= 0;
		    bool LOCAL= true;
		    bool CONSTANT= false;
		    if (CONSTANT) {
		      pparent[j]= 0.5;  // va b.
		     // pparent[j]= 1 / double(nC);
		    }
		    else
		    if (LOCAL) {
		      and_tree *Par= (and_tree *)OT.Parent;
          if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
            pparent[j]= (Par)->SmoothDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
            pparent[j]= (float)(Par)->ClassDistribution[j] / (float)Par->Cardinality;
		    }
		    else {  // Global. Everything computed wrt. global distribution
		      pparent[j]= Train.ClassDistribution()[j] / (float)Train.Cardinality();
		    //  pparent[j]*= nC / 2.0;  NO VA B
		    //  pparent[j] = sqrt(pparent[j]);  NO VA B
		
		    }
		    Mean+= pparent[j];
		  }
		
		  Mean/= nC;

  	  for(int j=0; j < nC; j++) {
       // pparent[j]/= 2 * Mean;  // NO va molt b

         pparent[j]= (pparent[j] + 0.5) / 2.0;     // VA PROU B

    //     pparent[j]= sqrt(pparent[j] * 0.5);         // va un poc pitjor per tamb va b
      }
		
      list<tree *>::iterator It, ItEnd;


      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> q(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          int j;
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
		      for(j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              mu[j]+= pNode * (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              mu[j]+= pNode * (float)Child->ClassDistribution[j] / (float)Card;
          }
        }
      }
/*
      for(int j=0; j < nC; j++) {
        mu[j]/= NChildren;
      }
*/

      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> q(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          int j;
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
		      for(j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;

//            SumSq[j]+= pNode * Sqr(pparent[j] - p[j]);  // ESTA ES LA BONA

            // NECESSARI SMOOTHING
        //     SumSq[j]+= pNode * sqrt(fabs((pparent[j] - p[j]) / (pparent[j] + p[j])));  va fatal


//            SumSq[j]+= Sqr(pparent[j] - p[j]);


//          SumSq[j]+= pNode * fabs(pow(mu[j] - p[j], 2));  //REGULAR

       //     double mu_without_itself= mu[j] - pNode * p[j];
        //    SumSq[j]+= pNode * fabs(pow(mu_without_itself - p[j], 2));

            SumSq[j]+= pNode * fabs(pow(pparent[j] - p[j], 3));  // esta pareix millor

            // NECESSARI SMOOTHING
//            double Logg= log(pparent[j] / p[j]);
//            SumSq[j]+= pNode * fabs(Logg);


       //      SumSq[j]+= pNode * fabs(pow(pparent[j] - p[j], 4));
          }
        }
      }

      double AcVar= 0;
      int j;
		  for(j=0; j < nC; j++) {
		    AcVar+= SumSq[j];  // ignoring class distribution
		   // AcVar+= pparent[j] * SumSq[j];  // taking into account class distribution
		                                    // FOR TWO CLASSES IS THE SAME, SINCE SumSq[0] = SumSq[1]
		
		    // Var = sqrt(Var); // sumar desviacions en compte de les varianes
      }

      // Children correction
		  // AcVar = AcVar / double(NChildren); //   / double(NChildren * (NChildren - 1.0));
		  // AcVar = AcVar * double(NChildren); //   / double(NChildren * (NChildren - 1.0));

		
      OT.Split_Optimality= AcVar;  // more difference better
    }

    else if ((Options.SSC==MGINI_SPLIT) || (Options.SSC == DKM)){
      int nC= Train.NumClasses();
      double MGINI=0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        vector<double> q(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          int j;
		      for(j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
           // q[j]= (float)Child->ClassDistribution[j] / (float)Card;
            q[j]= p[j];
          }
          //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
           double ValNode= 0;
          for(j=0; j < nC; j++) {
            for(int k=0; k < nC; k++) {
              if (j!=k) {
                double Prod= p[j]*q[k];
                if (Options.SSC == DKM)
                  Prod= sqrt(Prod);
                ValNode+= Prod;
              }
            }
          }
          MGINI+= pNode * ValNode;
        }
      }

      OT.Split_Optimality= - MGINI;
    }
    else if (Options.SSC==GENENTROPY_SPLIT) {
      int nC= Train.NumClasses();
      double GENENT=0;

      int NumChildren= OT.Children.size();

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          double NodeVal=0;
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
       //     NodeVal+= sqrt(p[j]);
            NodeVal+= pow(p[j], 0.5);
          }

       //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
        double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

          GENENT+= pNode * NodeVal;
        }
      }

       OT.Split_Optimality= - GENENT;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }
    else if (Options.SSC==SQDIF_SPLIT) {
      int nC= Train.NumClasses();
      double SQDIF=0;

      int NumChildren= OT.Children.size();

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
          }
          double p0= p[0];
          double p1= p[1];
          //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
           SQDIF+= pNode * Sqr(p0-p1);
        }
      }

      OT.Split_Optimality= SQDIF;

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }
    else if (Options.SSC==ROCV_SPLIT) {
      int nC= Train.NumClasses();
      double GENENT=0;
      vector<double> NCard;
      vector<vector<double> > pmat;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        long Card= Child->Cardinality;

        vector<double> p(nC);

        if (Card != 0) {
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
          }
        }
        //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
        double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
        NCard.push_back(pNode);

        pmat.push_back(p);
      }

      int NNodes= OT.Children.size();
      for(int k=0; k<NNodes; k++) {
        for(int l=0; l<NNodes; l++) {
          for(int i=0; i<nC; i++) {
            for(int j=0; j<nC; j++) {
              if (i<j) {
                GENENT+= NCard[k]*NCard[l]*pmat[k][i]*pmat[l][j];
              }
            }
          }
        }
      }
/*
      int NNodes= OT.Children.size();
        for(int l=0; l<NNodes; l++) {
          for(int i=0; i<nC; i++) {
            for(int j=0; j<nC; j++) {
              if (i<j) {
                GENENT+= NCard[l]*NCard[l]*pmat[l][i]*pmat[l][j];
              }
            }
          }
        }
*/
      OT.Split_Optimality=  +GENENT;

      int NumChildren= OT.Children.size();

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }
    else if (Options.SSC==ONE_POINT_LOCAL_ROC_AREA) {
       double Area;

       vector<double> Confus(4);  // ONLY FOR TWO CLASSES. ALTHOUGH IT CAN BE GENERALISED

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        if (Child->Cardinality != 0.0) {
          if (Child->PredictedClass == 0) {
            Confus[0]+= Child->SmoothDistribution[0] * Child->Cardinality;
            Confus[1]+= Child->SmoothDistribution[1] * Child->Cardinality;
          }
          else {
            Confus[2]+= Child->SmoothDistribution[0] * Child->Cardinality;
            Confus[3]+= Child->SmoothDistribution[1] * Child->Cardinality;
          }
        }
      }

       vector<vector<double> > Points;
       vector<double> Point(2);
       Point[0]= 0;
       Point[1]= 0;
       Points.push_back(Point);

       double c0= Confus[1]+Confus[3];
       double c1= Confus[0]+Confus[2];

       if (c0 == 0)
         Point[0]= 0.5;  // this never happens
       else
         Point[0]= Confus[1] / c0;

       if (c1 == 0)
         Point[1]= 0.5;  // this never happens
       else
         Point[1]= Confus[0] / c1;
       Points.push_back(Point);

       Point[0]= 1.0;
       Point[1]= 1.0;
       Points.push_back(Point);


//       compute_ROC_points(Vector, Points, ROC_WITH_TRAIN, Options);
       compute_ROC_AREA(Points, Area);
   //    cout << "ROC Area: " << Area << endl;
       OT.Split_Optimality= Area;

      int NumChildren= OT.Children.size();

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }


/*   // implemented but results are not good...
    if (true) {
    // else if ((Options.SSC==MACRO_AVERAGE_SPLIT) || (Options.SSC==MACRO_GEOMEAN_SPLIT)) {

      int nC= Train.NumClasses();
      vector<double> PartialAcc(nC);

      for(int j=0; j<nC; j++) {
        PartialAcc[j]= 0.0;
      }


      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
        double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

        vector<double> ProbTrain(nC);
        double Max= -1;
        for(int j=0; j<nC; j++) {
          if (Card == 0) {  // We assume equal probability for all the classes
            ProbTrain[j] = 1 / double(nC);
          }
          else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain[j]= (float)Child->SmoothDistribution[j];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain[j]= (float)Child->ClassDistribution[j] / (float)Card;
          }

          // double Acc= 1 -(ProbTrain * (1-ProbTrain)); // probability of assigning class a multiplied by probability of being of other class a
          // This way is nonsense for only two classes, since all Errors are the same
          if (Max < ProbTrain[j]) {
            Max= ProbTrain[j];
          }
        }

        double Acc;
        for(int j=0; j<nC; j++) {
          if (ProbTrain[j] < Max) {  // This doesn't seem to be the class selected for the node and there will be an error
            Acc= 0;
          }
          else {  // There can be more than one max, but this is not taken into account
            Acc= ProbTrain[j];
          }

          PartialAcc[j]+= pNode * Acc;
        }
      }

      double AverageAcc= 0.0;
      if (false) {
      // if (Options.SSC==MACRO_AVERAGE_SPLIT)  {

        for(int j=0; j<nC; j++) {
          AverageAcc+= PartialAcc[j];
        }
        AverageAcc/= nC;
      }
     if (true) {
//      else if (Options.SSC==MACRO_GEOMEAN_SPLIT) {
        for(int j=0; j<nC; j++) {
          AverageAcc*= PartialAcc[j];
        }
        AverageAcc= pow(AverageAcc, 1/double(nC));
      }

      OT.Split_Optimality= AverageAcc;
    }
*/
    else if (Options.SSC==MIXTURE_SPLIT) {
      // Mixture between MSE and Logloss

      double MSE=0;
      int nC= Train.NumClasses();
      double TotalTestValue= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        double MSE0= 0;

        for(int j=0; j<nC; j++) {
          long TestValue;
          TestValue= Child->ClassDistribution[j];
/*
          TestValue++; // DOUBLE SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + nC);
*/
/*
          const int K= 1;
          TestValue+= K; // K SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + K);
*/
          // WITHOUT DOUBLE SMOOTHING
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card);

          for(int k=0; k<nC; k++) {

            double ProbTrain;
            if (Card == 0) {  // We assume equal probability for all the classes
              ProbTrain = 1 / double(nC);
            }
            else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain= (float)Child->SmoothDistribution[k];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain= (float)Child->ClassDistribution[k] / (float)Card;
            }

            if (j!=k) {
              MSE0+= pTestValue * Sqr(0-ProbTrain);
            }
            else {
              MSE0+= pTestValue * Sqr(1-ProbTrain);
            }
          }
          TotalTestValue+= TestValue;
        }

        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
         double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

        MSE+= pNode * MSE0;

      }

      MSE/= /* TotalTestValue * */ nC;

      double MSE2;

      vector<and_tree *> Vector;

//      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      MSE2= compute_LogLoss(Vector, Options, MSE_WITH_TRAIN);
      MSE2= 1 - MSE2;

    //  cout << endl << endl << MSE << endl << MSE2 << endl;
    //  exit(0);

      OT.Split_Optimality= ((1-MSE) + MSE2) / 2.0;

      int NumChildren= OT.Children.size();

      Split_Children_Correction(OT, Train, Options, nC, NumChildren);
/*
      if (Options.SplittingCriteriaChildrenCorrection) {
        OT.Split_Optimality /= pow(NumChildren, Options.SplittingCriteriaChildrenCorrectionFactor);
      }
*/
    }


    else {
      cout << "Unknown Splitting criterion in function Compute_Criteria_OR" << endl;
    }


  if (Options.Components_In_Split_Criterion.Met == ADAPTATIVE_COMPONENT) {
     //     Options.Components.Show();
     //     Options.AccComponents.Show();
    // cout << "En este moment s'haurien de fer les coses b FISTRO" << endl;

     and_tree *a1Pa;
     a1Pa= (and_tree*)(OT.Children.front());

     float v1= Options.AccComponents.AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));

/*
     if (v1 == 0)
       cout << "matrix zeros FISTRO en criteria.cpp" << endl;
*/

     float f1= pow(2, - v1 * Options.Components_In_Split_Criterion.Factor);

     OT.Split_Optimality= OT.Split_Optimality * f1;
  }
  else if (Options.Components_In_Split_Criterion.Met == RANDOM_COMPONENTS) {
     //     Options.Components.Show();
     //     Options.AccComponents.Show();
   //  cout << "En este moment s'haurien de fer les coses b FISTRO" << endl;

     and_tree *a1Pa;
     a1Pa= (and_tree*)(OT.Children.front());

     float v1= Options.RandomComponents.AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));

/*
     if (v1 == 0)
       cout << "matrix zeros FISTRO en criteria.cpp" << endl;
*/

     float f1= pow(2, - v1 * Options.Components_In_Split_Criterion.Factor);
/*
     if (f1 != 1)
       cout << "iep" << endl;
*/
     OT.Split_Optimality= OT.Split_Optimality * f1;
  }



  if (Options.TestCostMethod != NO_TEST_COSTS) {

    if (Options.TestCostUse == TEST_COSTS_NO_USE) {
      // Nothing to do
    }
    else {
      and_tree *a1;
      a1= (and_tree*)(OT.Children.front());
      int Arg= a1->ActualArg;
      double TestCost= Options.TestCostVector[Arg];
      if (Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITHOUT_REPETITION) {
        or_tree *o2 = &OT;
        while(true) {
          and_tree *a2= (and_tree *)o2->Parent;

          if (a2->Parent == NULL)  // If it is the root. bye. The root does not have condition.
            break;

          if (a2->ActualArg == Arg) {  // this argument has already been used. We don't have to take it into account.
            TestCost= 0;
            break;
          }

          o2= (or_tree *)a2->Parent;
        }
      }


      // else if (Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITH_REPETITION) {
        // NO CONTROLLEM SI EIXE ARGUMENT JA S'HA UTILITZAT DALT
      double Denom= pow(TestCost + 1, Options.TestCostRelevanceInSplitting);

      OT.Split_Optimality= OT.Split_Optimality / Denom;

      // }
    }
  }





    #ifdef TREE_DEBUG
       cout << "Optimality of Node OR #" << OT.Num << " computed: " << OT.Split_Optimality << endl;
    #endif

 }


