/***************************************************************************
                          criteria.h  -  description
                             -------------------

    functionality        : for computing some criteria and defines structures related to these options
    relevant classes     : split_partition, split_selection_criteria ...
    relevant functions   :

    begin                : Mon Oct 22 2001
    copyright            : (C) 2001-2002 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//#include "estructures.h"
#ifndef CRITERIA_H
#define CRITERIA_H

#include <iostream>
using std::cout; // This is because some compilers...
using std::endl;
using namespace std;


#define NUM_PARTITIONS 3

//                   PARTITIONS

enum split_partition { PARTITION_MANY_CONST_EQ,               // Partition X=a, X=b, X=c...
               PARTITION_ONE_CONST_EQ,                // Partition X=a, X!=a
               PARTITION_ONE_VARIABLE_EQ,             // Partition X=Y, X!=Y
               PARTITION_ONE_CONST_DISEQ,              // Partition X<c, X>=c
               PARTITION_ONE_CONST_DISEQ_WITH_UNKNOWN // Partition X<c, X>=c, X=?
             } ;

//                    SPLIT SELECTION CRITERIA

enum split_selection_criteria { LEFT_FIRST,
                                GAIN, GAIN_RATIO, C45,
                                ADJ_GAIN, ADJ_GAIN_RATIO, C50,
                                SPLIT_EXPECTED_ERROR,
                                WEIGHTED_C45,
                                CART, DKM,
                                DESC_MDL, WEIGHTED_DESC_MDL,
																LOCAL_ROC_AREA,
																ONE_POINT_LOCAL_ROC_AREA,
																MSE_SPLIT, LOGLOSS_SPLIT,
																MCE_SPLIT,
																MSEE_SPLIT,
																CHILDVSPARENT_SPLIT,
																CHILDVSCHILD_SPLIT,
																AUCH_SPLIT,
																AUCS_SPLIT,
																SQDIF_SPLIT,
																MGINI_SPLIT,
																GENENTROPY_SPLIT,
																ROCV_SPLIT,
																MIXTURE_SPLIT
                               } ;

struct and_gain_struct {
  float InfoE;
};

struct or_gain_struct {
  float GainET;
  float SplitET;
  float GainRatioET;
};

struct and_cart_struct {
  float GiniE;
};

struct or_cart_struct {
  float GainET;
};

struct and_MDL_struct {
  double cost;
};

struct or_MDL_struct {
  float cost;
};

//typedef struct {
union and_criteria_structure {
  and_gain_struct Gain;
  and_cart_struct Cart;
  and_MDL_struct GMDL;
//};
 //tindex taula_cost;  // xapua temporal, a es pot millorar
 } ;

struct or_accumulated_structure {
  // Estes dos no es poden ficar en una unio perque es gasten en postpruning
    float AccExpectedError;
    long AccNumRules;
    float TestCost; // esta ha de estar amb l'anterior perqu a vegades es gasten les dos
  union {
    float AccMDL;
    long AccCoverage;
  };
} ;

struct and_accumulated_structure {
    // Estes dos no es poden ficar en una unio perque es gasten en postpruning
    float AccExpectedError;
    long AccNumRules;
    float TestCost;
  union {
    float AccMDL;
    long AccCoverage;
  };
} ;

union or_criteria_structure {
  or_gain_struct Gain;
  or_cart_struct Cart;
  or_MDL_struct GMDL ;
} ;


class and_tree;
class dataset;
class options;
// Computes the different criteria according to the values: SSC, ...
void Compute_Criteria_AND(and_tree &AT, const dataset &Train, const options &Options ) ;

class or_tree;
// Computes different criteria according to values: SSC, ...
void Compute_Criteria_OR(or_tree &OT, const dataset &Train, const options &Options ) ;

// corrects some criteria (no. of children).
void Split_Children_Correction(or_tree &OT, const dataset &Train, const options &Options, long nC, double NumChildren);


///////////////                     SECOND TREE CRITERION



enum second_tree_method{ SPLIT_OPTIMALITY, OPTIMALITY_RIVAL_RATIO,
                         OPTIMALITY_RIVAL_RATIO_DEPTH,OPTIMALITY_RIVAL_RATIO_COMPONENT,
                         OPTIMALITY_RIVAL_RATIO_COMPONENT_RANDOM,
                         SECOND_TOPMOST, SECOND_BOTTOMMOST,
                         SECOND_FRONT, SECOND_BACK,
                         SECOND_RANDOM, SECOND_RANDOM_DEPTH,
                         ROC_BISECTRIX } ;

class second_tree_criterion{
public:
  second_tree_method Met;
  float Factor;

public:
  second_tree_criterion(){
  }

  second_tree_criterion(second_tree_method Mt, float i=0) {
    Met=Mt;
    Factor=i;
  }
};



///////////////                      BEST TREE CRITERION
enum best_tree_criterion{ OCCAM_BEST,TEST_COST_BEST, OCCAM_AND_TEST_COST_BEST,
                          COVERAGE_BEST,CROSS_COVERAGE, ACCURACY_FALSE_TEST_BEST,
                          MDL_BEST, EXPECTED_ERROR_BEST, SPLIT_OPTIMALITY_BEST  }  ;



////////////////                     NUMERIC INTERVALS

enum numerical_interval_method { NO_LIMIT, MAX, LOG } ;

class numerical_interval_criterion {
public:
  numerical_interval_method Met;
  int Param;
  float Base;
public:
  numerical_interval_criterion() {
  }

  numerical_interval_criterion(numerical_interval_method M, int i=7, float Bas= 2.0) {
    Met= M;
    Param= i;
    Base= Bas;
  }
} ;

//////////////////                    PRE-PRUNING

enum  pruning_method { NO_PRUNING, PROPORTIONAL, EXPECTED_ERROR_PRUNING, MDL_PRUNING, STUMP_PRUNING,
                       PEP_PREPRUNING, OPTIMALITY_PREPRUNING, CARDINALITY_PREPRUNING } ;



class pruning_criterion {
public:
 // union {
    struct {
      float Strictness;
      long MaxUnclassified;
    } proportional;
    struct {
       float Factor;
    } ExpectedError;
    struct {
       float Factor;
    }  MDL;
    struct {
      int MaxDepth;
    } Stump;
// } ;

public:
  pruning_method Met;

public:
  pruning_criterion() {
  }

  pruning_criterion(pruning_method PM, float Strict=2.0, long MaxUnClass=10) {
    Met= PM;
    if (Met == NO_PRUNING) {
      // Nothing to initialise
    }
    else if (Met == PROPORTIONAL) {
      proportional.Strictness= Strict;
      proportional.MaxUnclassified = MaxUnClass;
    }
    else if (Met == EXPECTED_ERROR_PRUNING) {
      ExpectedError.Factor= Strict;
    }
    else if (Met == MDL_PRUNING) {
      MDL.Factor= Strict;
    }
    else if (Met == STUMP_PRUNING) {
      Stump.MaxDepth= (int)Strict;
    }
    else if (Met == PEP_PREPRUNING) {
      // Nothing to do
    }
  }

  bool To_Be_Pruned(float Error, int NumClasses, long MisClassified) {
    if (Met == NO_PRUNING) {
      return false;
    }
    else if (Met == PROPORTIONAL) {
       return ((Error < ((NumClasses-1)/proportional.Strictness) /float(NumClasses))
                && (MisClassified < proportional.MaxUnclassified)) ;
    }
  						
    else {
      cout << "Unknown pruning method" << endl;
      return false;
    }
  }
};




//////////////////                    POST-PRUNING

enum  post_pruning_method { NO_POSTPRUNING, PEP_POSTPRUNING } ;


#endif

