/***************************************************************************
                          exemples.cpp  -  description
                             -------------------
    functionality        : for the implementation of "dataset" class
    relevant classes     :
    relevant functions   :
    begin                : Fri Dec 27 2002
    copyright            : (C) 2002 by Cesar Ferri, Jose Hernandez, M.Jose Ramirez
    email                : {cferri,jorallo,mramirez}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "exemples.h"

#include "options.h"

dataset::dataset(const dataset &dtrain, const unsigned long NExamples, const options &Options) { // generates a new random dataset without touching the dtrain
	  num_arguments=dtrain.num_arguments;
    size_example=dtrain.size_example;
 	  argoffset=dtrain.argoffset;
	  exetable=dtrain.exetable;// taula amb els literals dels arguments
    profile=dtrain.profile;// perfil de la funcio
    vdist=dtrain.vdist; // distribuci de classes
    names=dtrain.names; // names of the arguments and classe optional
    weights=dtrain.weights;// weights of the values of classes


    unsigned long TrainSize= dtrain.data.size();
    if (Options.Comb2Single.InventedDatasetMethod == TRAINING_DISTRIBUTION) {  // training distribution
      for (unsigned long i=0; i < NExamples; i++) {
        example *Ex = new example(size_example);
        //  Ex.v = new byte(size_example);

        for (unsigned long j=0; j < num_arguments; j++) {
          argument Arg;

          long k= rand();
          k= k % TrainSize;

          // Arg.Value_d= 1;
          Arg= dtrain.Nth_arg(k, j);

          Set_Nth_arg(*Ex ,j, Arg);
        }
        data.push_back(*Ex);
        // delete [] Ex.v; // NO ES FA PERQUE ES QUEDA EN DATA
        // PERO SI S'HA DE FER
        delete Ex;
   	
        #ifdef VERBOSE
          cout << "Invented Example #" << i << "\r";
          cout.flush();
        #endif
 	    }
 	  }
 	  else {   // uniform distribution  // (Options.Comb2Single.InventedDatasetMethod == UNIFORM_DISTRIBUTION)
 	    vector<double> Min(num_arguments);
 	    vector<double> Max(num_arguments);

 	    unsigned long i, j;
      for (j=0; j < num_arguments; j++) {
 	      Min[j]= -1;
 	      Max[j]= -1;
 	    }
 	
 	    for (j=0; j < num_arguments; j++) {
        for (i=0; i < TrainSize; i++) {
          argument Arg;

          Arg= dtrain.Nth_arg(i, j);

          double d;

          kind Kind;

          Kind=profile.kinds[j];

          if ((Kind == NUMERIC_KIND) || (Kind == NUMERIC_WITH_UNKNOWN_KIND))
            d= (numeric)Arg.Value_n;
          else if (Kind == DISCRETE_KIND)
            d= (discrete)Arg.Value_d;
          else if (Kind == VARIABLE_KIND) {
            cout << "Error in dataset constructor" << endl; exit(-1);
          }
          else if (Kind == IGNORE_KIND) {
            // cout << "Error" << endl; exit(-1);
            // NOTHING TO DO
          }
          else cout << "Error: unknown argument kind in function Nth_arg";

          if ((Min[j] == -1) || (Min[j] > d)) {
            Min[j]= d;
          }

          if ((Max[j] == -1) || (Max[j] < d)) {
            Max[j]= d;
          }

        }
      }
      for (i=0; i < NExamples; i++) {
        example *Ex = new example(size_example);
        //  Ex.v = new byte(size_example);

        for (unsigned long j=0; j < num_arguments; j++) {
          argument Arg;

          kind Kind;

          Kind=profile.kinds[j];

          if ((Kind == NUMERIC_KIND) || (Kind == NUMERIC_WITH_UNKNOWN_KIND)) {
            double k= RandUniform(Min[j], Max[j]);
            Arg.Value_n = (numeric)k;
          }
          else if (Kind == DISCRETE_KIND) {
            double k= RandUniform(Min[j], Max[j] + 1);
            k = (discrete)k;
            if (k > Max[j]) k= Max[j];
            if (k < Min[j]) k= Min[j];
            Arg.Value_d = k;
          }
          else if (Kind == VARIABLE_KIND) {
            cout << "Error in dataset constructor" << endl; exit(-1);
          }
          else if (Kind == IGNORE_KIND) {
            // cout << "Error" << endl; exit(-1);
            // NOTHING TO DO
          }
          else cout << "Error: unknown argument kind in function Nth_arg";

          Set_Nth_arg(*Ex ,j, Arg);
        }
        data.push_back(*Ex);
        // delete [] Ex.v; // NO ES FA PERQUE ES QUEDA EN DATA
        // PERO SI S'HA DE FER
        delete Ex;
   	
        #ifdef VERBOSE
          cout << "Invented Example #" << i << "\r";
          cout.flush();
        #endif
      }

 	  }
 	
 	 // cout << endl;
 	 // cout << "SACABAT" << endl;
 	
 	  Setvdist();// actualitza la distribuci de classes
 	
 	  // cout << "SACABAT2" << endl;
  }




  void dataset::Append(const dataset &Second) {  // appends the examples of the second dataset to the first dataset
                                  // the data is copied completely (and deeply), and the original dataset unaffected.


    unsigned long SecondSize= Second.data.size();
    for (unsigned long i=0; i < SecondSize; i++) {
      example *Ex = new example(size_example);
      //  Ex->v = new byte(size_example);

      // Ex = Second.data[i];  // FA SHALLOW COPY!!! NO ES POT FER...
      for (unsigned long j=0; j < size_example; j++) {  // Deep copy
        (Ex->v)[j]= (Second.data[i]).v[j];
      }

      data.push_back(*Ex);

      // delete [] Ex.v; // NO ES FA PERQUE ES QUEDA EN DATA
      // PERO SI S'HA DE FER
      delete Ex;
   	}

 	  Setvdist();// actualitza la distribuci de classes i el majority class
  }



  void dataset::OutputOnFile2(const char *File, bool Open, bool Close, FILE * &F) const {
    int j;
    word4 i;

    if (Open) {
      F= fopen(File, "w");
      fprintf(F, "!TYPES:");
	
      int t= 1;
	    for (j=0;j<num_arguments;j++) {
	      if ((profile.kinds[j]==NUMERIC_KIND) || (profile.kinds[j]==NUMERIC_WITH_UNKNOWN_KIND)) {
	        fprintf(F, "0");
	      }
	      else
	        fprintf(F, "%d", t++);
	      if ((j + 1) < num_arguments) {
	        fprintf(F, ",");
	      }
	    }
    }

    for (i=0;i<data.size();i++) {
	    const example &ex =data[i];
	
	    fprintf(F, "\n");
	    for (j=0;j<num_arguments;j++) {
	      if ((profile.kinds[j]==NUMERIC_KIND) || (profile.kinds[j]==NUMERIC_WITH_UNKNOWN_KIND))
 		      fprintf(F, "%f",*((numeric *)(ex.v+argoffset[j])));
 		    else if (profile.kinds[j]==IGNORE_KIND)
 		      fprintf(F, "IGNORED");
	      else {
	        discrete d;
	        d= ex.v[argoffset[j]];
	        string s= NameArgumentValue(j, d);
		      fprintf(F, "%s", s.c_str());
		    }
	      if ((j + 1) < num_arguments) {
	        fprintf(F, ",");
	      }
      }
	  }
	
	  if (Close) {
		  fprintf(F, "\n");
	    fclose(F);
    }
  }
